import cv2
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torchvision import models, transforms, datasets
import torchvision.transforms.functional as TF
import os
import torch.optim as optim
import torch.nn.init as init
from loguru import logger

__all__ = ['Classifier']

MODELS = {
    'resetnet18': models.resnet18,
    'squeezenet1_0': models.squeezenet1_0,
    'swin_t': models.swin_t,
    'swin_s': models.swin_s,
    'swin_b': models.swin_b
}

WEIGHTS = {
    'resetnet18': models.ResNet18_Weights.DEFAULT,
    'squeezenet1_0': models.SqueezeNet1_0_Weights.DEFAULT,
    'swin_t': models.Swin_T_Weights.DEFAULT,
    'swin_s': models.Swin_S_Weights.DEFAULT,
    'swin_b': models.Swin_B_Weights.DEFAULT
}

OPTIMIZER = {
    'Adadelta': optim.Adadelta,
    'Adagrad': optim.Adagrad,
    'Adam': optim.Adam,
    'AdamW': optim.AdamW,
    'Adamax': optim.Adamax,
    'ASGD': optim.ASGD,
    'NAdam': optim.NAdam,
    'RAdam': optim.RAdam,
    'RMSprop': optim.RMSprop,
    'Rprop': optim.Rprop,
    'SGD': optim.SGD
}


def set_parameter_requires_grad(model, feature_extracting):
    if feature_extracting:
        for param in model.parameters():
            param.requires_grad = False


class Classifier():

    def __init__(self) -> None:
        self._mean = np.array([0.485, 0.456, 0.406])
        self._std = np.array([0.229, 0.224, 0.225])
        self._resize = (224, 224)
        self._device = torch.device(
            "cuda:0" if torch.cuda.is_available() else "cpu")

        # 训练相关
        self.feature_extract = True

    def _preprocess(self, bgr):
        rgb = cv2.cvtColor(bgr, cv2.COLOR_BGR2RGB)
        rgb = cv2.resize(rgb, self._resize)
        tensor = TF.to_tensor(rgb)
        tensor = TF.normalize(tensor, self._mean, self._std)
        tensor = torch.unsqueeze(tensor, 0)
        return tensor

    def get_avaliable_model_name(self):
        """获取当前支持的模型名称列表。

        Returns:
            list[str]: 模型名称列表
        """
        return list(MODELS.keys())

    def get_base_model(self, model_name='resetnet18', weights='pretrained'):
        """获取基础模型

        Args:
            model_name (str, optional): 模型名称。可用 `get_avaliable_model_name` 查看支持的模型。 Defaults to 'resetnet18'.
            weights (any, optional): 当为 'pretrained' 时，会载入训练好的参数，
            当为 None 时，会随机初始化参数，当为其它值时，会当作权重文件路径并载入相应的权重。 Defaults to 'pretrained'.

        Returns:
            torch.nn.Module: torch 模型
        """
        _model = MODELS[model_name]
        if weights == 'pretrained':
            w = WEIGHTS[model_name]
            model = _model(weights=w)
        else:
            model = _model(weights=None)
            if weights is not None:
                model.load_state_dict(weights)
        model = model.to(self._device)
        model.eval()
        self.model = model
        self.categories = models.SqueezeNet1_0_Weights.DEFAULT.meta[
            'categories']
        return model

    def create_model(self, base_model_name, num_classes, feature_extract=True):
        """创建分类模型

        Args:
            base_model_name (str): 基础模型名称，创建模型时会以此模型作为骨干网络。
            可用 `get_avaliable_model_name` 查看支持的模型。
            num_classes (int): 分类数量
            feature_extract (bool, optional): 为 True 时，进行特征提取，为 False 时，进行微调. Defaults to True.

        Returns:
            torch.nn.Module: torch 模型
        """
        _model = MODELS[base_model_name]
        w = WEIGHTS[base_model_name]
        model_ft = _model(weights=w)
        set_parameter_requires_grad(model_ft, feature_extract)
        if base_model_name == 'resetnet18':
            num_ftrs = model_ft.fc.in_features
            model_ft.fc = nn.Linear(num_ftrs, num_classes)
        elif base_model_name == 'squeezenet1_0':
            final_conv = nn.Conv2d(512,
                                   num_classes,
                                   kernel_size=(1, 1),
                                   stride=(1, 1))
            # see https://github.com/pytorch/vision/blob/v0.13.1/torchvision/models/squeezenet.py#L88
            init.normal_(final_conv.weight, mean=0, std=0.01)
            if final_conv.bias is not None:
                init.constant_(final_conv.bias, 0)
            model_ft.classifier[1] = final_conv
            model_ft.num_classes = num_classes
        elif base_model_name == 'swin_t' or base_model_name == 'swin_s' or base_model_name == 'swin_b':
            num_ftrs = model_ft.head.in_features
            model_ft.head = nn.Linear(num_ftrs, num_classes)
            # see https://github.com/pytorch/vision/blob/v0.13.1/torchvision/models/swin_transformer.py#L399
            init.trunc_normal_(model_ft.head.weight, std=0.02)
            if model_ft.head.bias is not None:
                init.zeros_(model_ft.head.bias)
        model_ft = model_ft.to(self._device)
        self.model = model_ft
        self.feature_extract = feature_extract
        return model_ft

    def load_model(self, path):
        """载入模型

        Args:
            path (str): 模型文件路径，文件类型一般是 pth 
        """
        self.model = torch.load(path).to(self._device)

    def load_label(self, label_path):
        """载入标签，该标签会在解析预测结果时用到

        Args:
            label_path (str): 标签文件路径，一般来说，使用 `create_dataloaders` 方法
            生成的标签文件
        """
        labels = []
        with open(label_path, 'r', encoding='utf-8') as f:
            for line in f:
                labels.append(line.strip())

        self.categories = labels

    def save_model(self, file_path='MyModel.pth'):
        """保存模型

        Args:
            file_path (str, optional): 模型文件保存路径. Defaults to 'MyModel.pth'.
        """
        torch.save(self.model, file_path)
        logger.debug(f'模型已保存到 {file_path}')

    def load_weights(self, weight_path):
        """载入权重

        Args:
            weight_path (str): 权重文件路径
        """
        self.model.load_state_dict(torch.load(weight_path))

    def predict(self, img):
        """运行推理，返回 softmax 后的结果

        Args:
            img (numpy.ndarray): BGR格式的图片
        
        Returns:
            numpy.ndarray: 返回 N 维的向量，N是分类数量。每一项对应相应分类的置信度。
        """
        self.model.eval()
        tensor = self._preprocess(img)
        with torch.no_grad():
            tensor = tensor.to(self._device)
            output = self.model(tensor)
            return nn.functional.softmax(output, dim=1).cpu().numpy()[0]

    def parse_prediction(self, predictions, top_k=1):
        """解析模型的推理结果。在使用前，请确保你已载入相应的标签文件。
           如果你直接使用基础模型（ImageNet 1000分类），则无需载入标签文件。

        Args:
            predictions (numpy.ndarray): `predict` 方法返回的结果
            top_k (int, optional): 解析可能性最高的 k 个结果. Defaults to 1.
        """
        idx_score_pairs = self.get_top_k(predictions, top_k)
        print('该图片')
        for t in idx_score_pairs:
            print(f"有{t[1]*100:.1f}%的可能是'{self.categories[t[0]]}'")

    def get_top_k(self, predictions, top_k=1):
        """获取分数最高的 `top_k` 个结果

        Args:
            predictions (numpy.ndarray): `predict` 方法返回的结果
            top_k (int, optional): 该数值决定返回的结果个数. Defaults to 1.

        Returns:
            list(tuple): 每个元组都是一个“索引-分数”对，元组内的“分数”就是 `predictions` 中
            成绩最高的 `top_k` 个之一，而“索引”就是该“分数”在 `predictions` 的索引号。
        """
        if top_k > len(predictions):
            top_k = len(predictions)
        top_k_idx = np.argpartition(predictions, -top_k)[-top_k:]
        scores = [predictions[i] for i in top_k_idx]
        return list(zip(list(top_k_idx), scores))

    def create_dataloaders(self,
                           dir,
                           label_save_path='labels.txt',
                           batch_size=4,
                           random_hflips=True,
                           color_jitter=(0.2, 0.2, 0.1, 0.1),
                           random_rotation=10,
                           random_crop=(0.8, 1.0)):
        """创建数据载入器

        Args:
            dir (str): 存放数据集的文件夹路径，该目录下应包含 train 和 val 两个文件夹
            label_save_path (str, optional): 标签文件保存路径. Defaults to 'labels.txt'.
            batch_size (int, optional): 批大小，根据自身电脑的内存或显存大小而设，数字越大，训练时占用的内存或显存就越大. Defaults to 4.
            random_hflips (bool, optional): 是否随机水平镜像图片. Defaults to True.
            color_jitter (tuple, optional): 随机改变图片的亮度、对比度、饱和度、色相，详细说明见 
                https://pytorch.org/vision/0.13/generated/torchvision.transforms.ColorJitter.html#torchvision.transforms.ColorJitter.
                Defaults to (0.2, 0.2, 0.1, 0.1).
            random_rotation (int, optional): 随机改变图片的角度. Defaults to 10.
            random_crop (tuple, optional): 随机剪裁图片的范围. Defaults to (0.8, 1.0).

        Returns:
            dict: 含有训练集与验证集的数据载入器
        """

        train_tf = [
            transforms.ColorJitter(*color_jitter),
            transforms.RandomRotation(random_rotation),
            transforms.RandomResizedCrop(224, scale=random_crop),
        ]
        if random_hflips:
            train_tf.append(transforms.RandomHorizontalFlip())

        train_tf.append(transforms.ToTensor())
        train_tf.append(
            transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225]))
        data_transforms = {
            'train':
            transforms.Compose(train_tf),
            'val':
            transforms.Compose([
                transforms.Resize(224),
                transforms.ToTensor(),
                transforms.Normalize([0.485, 0.456, 0.406],
                                     [0.229, 0.224, 0.225])
            ]),
        }

        image_datasets = {
            x: datasets.ImageFolder(os.path.join(dir, x), data_transforms[x])
            for x in ['train', 'val']
        }
        dataloaders = {
            x: torch.utils.data.DataLoader(image_datasets[x],
                                           batch_size=batch_size,
                                           shuffle=True,
                                           num_workers=0)
            for x in ['train', 'val']
        }
        class_names = image_datasets['train'].classes

        with open(label_save_path, 'w', encoding='utf-8') as f:
            for c in class_names:
                f.write(f'{c}\r\n')

        logger.debug(f'分类标签已保存到 {os.path.abspath(label_save_path)} 文件')
        return dataloaders

    def create_optimizer(self, optim_name, lr=0.001):
        params_to_update = self.model.parameters()
        print("需要训练的参数:")
        if self.feature_extract:
            params_to_update = []
            for name, param in self.model.named_parameters():
                if param.requires_grad == True:
                    params_to_update.append(param)
                    print("\t", name)
        else:
            for name, param in self.model.named_parameters():
                if param.requires_grad == True:
                    print("\t", name)
        optimizer = OPTIMIZER[optim_name](params_to_update, lr=lr)

        return optimizer

    def get_avaliable_optimizer_name(self):
        """获取可用的优化器名称

        Returns:
            list[str]: 优化器名称列表
        """
        return list(OPTIMIZER.keys())

    def train(self,
              dataloaders,
              optimizer='Adam',
              lr=0.001,
              save_path='.',
              num_epochs=10):
        """训练模型

        Args:
            dataloaders (dict): 由 `create_dataloaders` 方法创建的数据载入器
            optimizer (str, optional): 优化器名称，可用 `get_avaliable_optimizer_name` 查看支持的优化器。 Defaults to 'Adam'.
            lr (float, optional): 学习率. Defaults to 0.001.
            save_path (str, optional): 权重保存目录. Defaults to '.'.
            num_epochs (int, optional): 训练次数. Defaults to 10.
        """

        model = self.model.to(self._device)
        device = self._device
        optimizer = self.create_optimizer(optimizer, lr)

        train_loader = dataloaders['train']
        val_loader = dataloaders['val']

        best_acc = 0.

        for ep in range(num_epochs):
            model.train()
            train_loss = 0.0
            train_acc = 0.
            for images, labels in iter(train_loader):
                images = images.to(device)
                labels = labels.to(device)
                optimizer.zero_grad()
                outputs = model(images)

                loss = F.cross_entropy(outputs, labels)
                _, preds = torch.max(outputs, 1)
                train_acc += torch.sum(preds == labels.data)
                train_loss += loss.item()
                loss.backward()
                optimizer.step()
            train_loss /= len(train_loader)
            train_acc /= len(train_loader.dataset)

            model.eval()
            test_loss = 0.0
            test_acc = 0.
            for items in iter(val_loader):
                images, labels = items[:2]
                images = images.to(device)
                labels = labels.to(device)
                outputs = model(images)

                loss = F.cross_entropy(outputs, labels)
                _, preds = torch.max(outputs, 1)
                test_acc += torch.sum(preds == labels.data)
                test_loss += loss.item()

            test_loss /= len(val_loader)
            test_acc /= len(val_loader.dataset)
            logger.debug(
                f"epoch {ep} - train loss: {train_loss}, val loss: {test_loss}"
            )

            logger.debug(f"train acc: {train_acc}, val acc: {test_acc}")

            if test_acc > best_acc:
                torch.save(model.state_dict(),
                           os.path.join(save_path, 'best_ckpt.pth'))
                logger.debug(
                    f"权重已保存到 {os.path.join(save_path,'best_ckpt.pth')}")
                best_acc = test_acc



import cv2.aruco as aruco
import numpy as np

default_cameraMatrix = np.array([[392.75533375, 0., 284.73851876],
                                 [0., 413.9101659, 254.38799118], [0., 0.,
                                                                   1.]])
default_distCoeffs = np.array(
    [[0.0290911, -1.24869976, -0.01714862, -0.02837211, 3.59261617]])


class ArucoTracker():
    def __init__(self,
                 car=None,
                 xp=-1.2,
                 zp=3,
                 xd=0,
                 zd=0,
                 target_distance=0.3,
                 output_limit=0.3,
                 marker_size=0.06,
                 cameraMatrix=default_cameraMatrix,
                 distCoeffs=default_distCoeffs) -> None:
        self.aruco_dict = aruco.custom_dictionary(10, 6)
        self.car = car
        self.xp = xp
        self.zp = zp
        self.xd = xd
        self.zd = zd
        self.target_distance = target_distance
        self.output_limit = output_limit
        self.marker_size = marker_size
        self.cameraMatrix = cameraMatrix
        self.distCoeffs = distCoeffs
        self._last_angle = 0
        self._last_z_error = 0

    def detect(self, bgr):
        corners, ids, _ = aruco.detectMarkers(bgr, self.aruco_dict)
        if ids is not None:
            aruco.drawDetectedMarkers(bgr, corners, ids)
        return corners, ids

    def track(self, bgr):
        left_speed = 0
        right_speed = 0
        corners, ids, _ = aruco.detectMarkers(bgr, self.aruco_dict)
        if ids is not None:
            _, tvecs, _ = aruco.estimatePoseSingleMarkers(
                corners, self.marker_size, self.cameraMatrix, self.distCoeffs)

            x_error = tvecs[0][0][0].item()
            z_error = tvecs[0][0][2].item() - self.target_distance
            angle = np.arctan2(x_error, tvecs[0][0][2].item())

            if tvecs[0][0][2] > 0.05:
                aruco.drawDetectedMarkers(bgr, np.array([corners[0]]),
                                          np.array([ids[0]]))
                x_output = self.xp * angle + (angle -
                                              self._last_angle) * self.xd
                z_output = self.zp * z_error + (z_error -
                                                self._last_z_error) * self.zd
                self._last_angle = x_error
                self._last_z_error = z_error
                left_speed = z_output - x_output
                right_speed = z_output + x_output
                left_speed = max(min(left_speed, self.output_limit),
                                 -self.output_limit)
                right_speed = max(min(right_speed, self.output_limit),
                                  -self.output_limit)
            else:
                self._last_angle = 0
                self._last_z_error = 0
        else:
            self._last_angle = 0
            self._last_z_error = 0

        if self.car is not None:
            self.car.set_motors(left_speed, right_speed)



from PIL import Image
from fro_AI.face_evoLve.align.detector import detect_faces
from fro_AI.face_evoLve.align.align_trans import get_reference_facial_points, warp_and_crop_face
import numpy as np
import os
from tqdm import tqdm
import argparse
import glob

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description = "face alignment")
    parser.add_argument("-source_root", "--source_root", help = "specify your source dir", default = "./data/test", type = str)
    parser.add_argument("-dest_root", "--dest_root", help = "specify your destination dir", default = "./data/test_Aligned", type = str)
    parser.add_argument("-crop_size", "--crop_size", help = "specify size of aligned faces, align and crop with padding", default = 112, type = int)
    args = parser.parse_args()

    source_root = args.source_root # specify your source dir
    dest_root = args.dest_root # specify your destination dir
    crop_size = args.crop_size # specify size of aligned faces, align and crop with padding
    scale = crop_size / 112.
    reference = get_reference_facial_points(default_square = True) * scale

    cwd = os.getcwd() # delete '.DS_Store' existed in the source_root
    os.chdir(source_root)
    os.system("find . -name '*.DS_Store' -type f -delete")
    os.system("find . -name '*.ipynb_checkpoints' -prune -exec rm -rf {} \;")
    os.chdir(cwd)

    if not os.path.isdir(dest_root):
        os.mkdir(dest_root)

    for subfolder in tqdm(os.listdir(source_root)):
        if not os.path.isdir(os.path.join(dest_root, subfolder)):
            os.mkdir(os.path.join(dest_root, subfolder))
        for image_name in os.listdir(os.path.join(source_root, subfolder)):
            print("Processing\t{}".format(os.path.join(source_root, subfolder, image_name)))
            img = Image.open(os.path.join(source_root, subfolder, image_name))
            try: # Handle exception
                _, landmarks = detect_faces(img)
            except Exception as e:
                print("{} is discarded due to exception!".format(os.path.join(source_root, subfolder, image_name)))
                print(e)
                continue
            if len(landmarks) == 0: # If the landmarks cannot be detected, the img will be discarded
                print("{} is discarded due to non-detected landmarks!".format(os.path.join(source_root, subfolder, image_name)))
                continue
            facial5points = [[landmarks[0][j], landmarks[0][j + 5]] for j in range(5)]
            warped_face = warp_and_crop_face(np.array(img), facial5points, reference, crop_size=(crop_size, crop_size))
            img_warped = Image.fromarray(warped_face)
            if image_name.split('.')[-1].lower() not in ['jpg', 'jpeg']: #not from jpg
                image_name = '.'.join(image_name.split('.')[:-1]) + '.jpg'
            img_warped.save(os.path.join(dest_root, subfolder, image_name))



import torch.nn as nn
from torch.nn import Linear, Conv2d, BatchNorm1d, BatchNorm2d, ReLU, Dropout, MaxPool2d, Sequential, Module


# Support: ['ResNet_50', 'ResNet_101', 'ResNet_152']


def conv3x3(in_planes, out_planes, stride = 1):
    """3x3 convolution with padding"""

    return Conv2d(in_planes, out_planes, kernel_size = 3, stride = stride,
                     padding = 1, bias = False)


def conv1x1(in_planes, out_planes, stride = 1):
    """1x1 convolution"""

    return Conv2d(in_planes, out_planes, kernel_size = 1, stride = stride, bias = False)


class BasicBlock(Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride = 1, downsample = None):
        super(BasicBlock, self).__init__()
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = BatchNorm2d(planes)
        self.relu = ReLU(inplace = True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = BatchNorm2d(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class Bottleneck(Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride = 1, downsample = None):
        super(Bottleneck, self).__init__()
        self.conv1 = conv1x1(inplanes, planes)
        self.bn1 = BatchNorm2d(planes)
        self.conv2 = conv3x3(planes, planes, stride)
        self.bn2 = BatchNorm2d(planes)
        self.conv3 = conv1x1(planes, planes * self.expansion)
        self.bn3 = BatchNorm2d(planes * self.expansion)
        self.relu = ReLU(inplace = True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class ResNet(Module):

    def __init__(self, input_size, block, layers, zero_init_residual = True):
        super(ResNet, self).__init__()
        assert input_size[0] in [112, 224], "input_size should be [112, 112] or [224, 224]"
        self.inplanes = 64
        self.conv1 = Conv2d(3, 64, kernel_size = 7, stride = 2, padding = 3, bias = False)
        self.bn1 = BatchNorm2d(64)
        self.relu = ReLU(inplace = True)
        self.maxpool = MaxPool2d(kernel_size = 3, stride = 2, padding = 1)
        self.layer1 = self._make_layer(block, 64, layers[0])
        self.layer2 = self._make_layer(block, 128, layers[1], stride = 2)
        self.layer3 = self._make_layer(block, 256, layers[2], stride = 2)
        self.layer4 = self._make_layer(block, 512, layers[3], stride = 2)

        self.bn_o1 = BatchNorm2d(2048)
        self.dropout = Dropout()
        if input_size[0] == 112:
            self.fc = Linear(2048 * 4 * 4, 512)
        else:
            self.fc = Linear(2048 * 8 * 8, 512)
        self.bn_o2 = BatchNorm1d(512)

        for m in self.modules():
            if isinstance(m, Conv2d):
                nn.init.kaiming_normal_(m.weight, mode = 'fan_out', nonlinearity = 'relu')
            elif isinstance(m, BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

        # Zero-initialize the last BN in each residual branch,
        # so that the residual branch starts with zeros, and each residual block behaves like an identity.
        # This improves the model by 0.2~0.3% according to https://arxiv.org/abs/1706.02677
        if zero_init_residual:
            for m in self.modules():
                if isinstance(m, Bottleneck):
                    nn.init.constant_(m.bn3.weight, 0)
                elif isinstance(m, BasicBlock):
                    nn.init.constant_(m.bn2.weight, 0)

    def _make_layer(self, block, planes, blocks, stride = 1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = Sequential(
                conv1x1(self.inplanes, planes * block.expansion, stride),
                BatchNorm2d(planes * block.expansion),
            )

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample))
        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layers.append(block(self.inplanes, planes))

        return Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.bn_o1(x)
        x = self.dropout(x)
        x = x.view(x.size(0), -1)
        x = self.fc(x)
        x = self.bn_o2(x)

        return x


def ResNet_50(input_size, **kwargs):
    """Constructs a ResNet-50 model.
    """
    model = ResNet(input_size, Bottleneck, [3, 4, 6, 3], **kwargs)

    return model


def ResNet_101(input_size, **kwargs):
    """Constructs a ResNet-101 model.
    """
    model = ResNet(input_size, Bottleneck, [3, 4, 23, 3], **kwargs)

    return model


def ResNet_152(input_size, **kwargs):
    """Constructs a ResNet-152 model.
    """
    model = ResNet(input_size, Bottleneck, [3, 8, 36, 3], **kwargs)

    return model
